/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_misc.c

#include "quakedef.h"

extern cvar_t		gl_normalMaps; //rww

extern void R_InitBubble();

/*
==================
R_InitTextures
==================
*/
void	R_InitTextures (void)
{
	int		x,y, m;
	byte	*dest;

// create a simple checkerboard texture for the default
	r_notexture_mip = Hunk_AllocName (sizeof(texture_t) + 16*16+8*8+4*4+2*2, "notexture");
	
	r_notexture_mip->width = r_notexture_mip->height = 16;
	r_notexture_mip->offsets[0] = sizeof(texture_t);
	r_notexture_mip->offsets[1] = r_notexture_mip->offsets[0] + 16*16;
	r_notexture_mip->offsets[2] = r_notexture_mip->offsets[1] + 8*8;
	r_notexture_mip->offsets[3] = r_notexture_mip->offsets[2] + 4*4;
	
	for (m=0 ; m<4 ; m++)
	{
		dest = (byte *)r_notexture_mip + r_notexture_mip->offsets[m];
		for (y=0 ; y< (16>>m) ; y++)
			for (x=0 ; x< (16>>m) ; x++)
			{
				if (  (y< (8>>m) ) ^ (x< (8>>m) ) )
					*dest++ = 0;
				else
					*dest++ = 0xff;
			}
	}	
}

byte	dottexture[8][8] =
{
	{0,1,1,0,0,0,0,0},
	{1,1,1,1,0,0,0,0},
	{1,1,1,1,0,0,0,0},
	{0,1,1,0,0,0,0,0},
	{0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0},
};
void R_InitParticleTexture (void)
{
	int		x,y;
	byte	data[8][8][4];

	//
	// particle texture
	//
	particletexture = texture_extension_number++;
    GL_Bind(particletexture);

	for (x=0 ; x<8 ; x++)
	{
		for (y=0 ; y<8 ; y++)
		{
			data[y][x][0] = 255;
			data[y][x][1] = 255;
			data[y][x][2] = 255;
			data[y][x][3] = dottexture[x][y]*255;
		}
	}
	glTexImage2D (GL_TEXTURE_2D, 0, gl_alpha_format, 8, 8, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);

	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
}

/*
===============
R_Envmap_f

Grab six views for environment mapping tests
===============
*/
void R_Envmap_f (void)
{
	byte	buffer[256*256*4];

	glDrawBuffer  (GL_FRONT);
	glReadBuffer  (GL_FRONT);
	envmap = true;

	r_refdef.vrect.x = 0;
	r_refdef.vrect.y = 0;
	r_refdef.vrect.width = 256;
	r_refdef.vrect.height = 256;

	r_refdef.viewangles[0] = 0;
	r_refdef.viewangles[1] = 0;
	r_refdef.viewangles[2] = 0;
	GL_BeginRendering (&glx, &gly, &glwidth, &glheight);
	R_RenderView ();
	glReadPixels (0, 0, 256, 256, GL_RGBA, GL_UNSIGNED_BYTE, buffer);
	COM_WriteFile ("env0.rgb", buffer, sizeof(buffer));		

	r_refdef.viewangles[1] = 90;
	GL_BeginRendering (&glx, &gly, &glwidth, &glheight);
	R_RenderView ();
	glReadPixels (0, 0, 256, 256, GL_RGBA, GL_UNSIGNED_BYTE, buffer);
	COM_WriteFile ("env1.rgb", buffer, sizeof(buffer));		

	r_refdef.viewangles[1] = 180;
	GL_BeginRendering (&glx, &gly, &glwidth, &glheight);
	R_RenderView ();
	glReadPixels (0, 0, 256, 256, GL_RGBA, GL_UNSIGNED_BYTE, buffer);
	COM_WriteFile ("env2.rgb", buffer, sizeof(buffer));		

	r_refdef.viewangles[1] = 270;
	GL_BeginRendering (&glx, &gly, &glwidth, &glheight);
	R_RenderView ();
	glReadPixels (0, 0, 256, 256, GL_RGBA, GL_UNSIGNED_BYTE, buffer);
	COM_WriteFile ("env3.rgb", buffer, sizeof(buffer));		

	r_refdef.viewangles[0] = -90;
	r_refdef.viewangles[1] = 0;
	GL_BeginRendering (&glx, &gly, &glwidth, &glheight);
	R_RenderView ();
	glReadPixels (0, 0, 256, 256, GL_RGBA, GL_UNSIGNED_BYTE, buffer);
	COM_WriteFile ("env4.rgb", buffer, sizeof(buffer));		

	r_refdef.viewangles[0] = 90;
	r_refdef.viewangles[1] = 0;
	GL_BeginRendering (&glx, &gly, &glwidth, &glheight);
	R_RenderView ();
	glReadPixels (0, 0, 256, 256, GL_RGBA, GL_UNSIGNED_BYTE, buffer);
	COM_WriteFile ("env5.rgb", buffer, sizeof(buffer));		

	envmap = false;
	glDrawBuffer  (GL_BACK);
	glReadBuffer  (GL_BACK);
	GL_EndRendering ();
}

//rww
qboolean r_hasInitialGamma = false;
unsigned short r_initialGamma[3][256];
qboolean r_hasCustomGamma = false;
float r_lastGamma = -1.0f;
extern HDC maindc;
void R_GammaInit(void)
{
	if (GetDeviceGammaRamp(maindc, r_initialGamma))
	{
		r_hasInitialGamma = true;
	}
}

//rww
void R_GammaShutdown(void)
{
	if (!r_hasCustomGamma)
	{
		return;
	}

	SetDeviceGammaRamp(maindc, r_initialGamma);
}

//rww
void R_UploadGamma(float gamma)
{
	unsigned short ngam[3][256];
	int i;
	for (i = 0; i < 256; i++)
	{
		byte c = 0;
		byte g;
		unsigned short r;
		if (gamma == 1.0f)
		{
			g = i;
		}
		else
		{
			float f = pow(((float)i)/255.0f, gamma)*255.0f;
			if (f < 0.0f)
			{
				f = 0.0f;
			}
			else if (f > 255.0f)
			{
				f = 255.0f;
			}
			g = (byte)f;
		}
		r = c|(g<<8);
		ngam[0][i] = r;
		ngam[1][i] = r;
		ngam[2][i] = r;
	}

	SetDeviceGammaRamp(maindc, ngam);
}

//rww
void R_GammaCheck(void)
{
	if (!r_hasInitialGamma)
	{
		return;
	}

	if (v_gamma.value != r_lastGamma)
	{
		r_hasCustomGamma = true;
		r_lastGamma = v_gamma.value;
		R_UploadGamma(v_gamma.value);
	}
}

//rww
void R_SwapInterval_f(void)
{
	static qboolean vsyncOn = true;
	if (!g_glExt.wglSwapIntervalEXT)
	{
		return;
	}

	if (vsyncOn)
	{
		g_glExt.wglSwapIntervalEXT(0);
		Con_Printf("vsync disabled.\n");
	}
	else
	{
		g_glExt.wglSwapIntervalEXT(1);
		Con_Printf("vsync enabled.\n");
	}

	vsyncOn = !vsyncOn;
}

//stupid bot waypoint hackery, trying to get the whole bot ai/nav system done in about 3 hours, sorry
typedef struct botWaypoint_s
{
	float		pos[3];
} botWaypoint_t;

#define MAX_BOT_WAYPOINTS	8192
botWaypoint_t *g_botWaypoints = NULL;
int g_numBotWaypoints = 0;

void R_DrawBotWP(void)
{
	int i;
	if (!developer.value || g_numBotWaypoints <= 0 || !g_botWaypoints)
	{
		return;
	}

	for (i = 0; i < g_numBotWaypoints; i++)
	{
		botWaypoint_t *wp = g_botWaypoints+i;
		glPointSize(5.0f);
		glDisable(GL_TEXTURE_2D);
		glColor4f(1.0f, 0.0f, 0.0f, 1.0f);
		glBegin(GL_POINTS);
			glVertex3fv(wp->pos);
		glEnd();
		if (i > 0)
		{
			glColor4f(0.0f, 1.0f, 0.0f, 1.0f);
			glBegin(GL_LINES);
				glVertex3fv(wp->pos);
				glVertex3fv((wp-1)->pos);
			glEnd();
		}
		glEnable(GL_TEXTURE_2D);
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
		glPointSize(1.0f);
	}
}

void R_AddBotWP_f(void)
{
	botWaypoint_t *wp;

	if (!developer.value)
	{
		return;
	}

	if (!cl.worldmodel)
	{
		return;
	}

	if (g_numBotWaypoints >= MAX_BOT_WAYPOINTS)
	{
		Con_Printf("Too many waypoints.\n");
		return;
	}

	if (!g_botWaypoints)
	{
		g_botWaypoints = malloc(sizeof(botWaypoint_t)*MAX_BOT_WAYPOINTS);
	}

	wp = g_botWaypoints+g_numBotWaypoints;
	g_numBotWaypoints++;

	VectorCopy(cl.simorg, wp->pos);
}

void R_RemBotWP_f(void)
{
	if (!developer.value)
	{
		return;
	}

	g_numBotWaypoints--;
	if (g_numBotWaypoints <= 0)
	{
		if (g_botWaypoints)
		{
			free(g_botWaypoints);
			g_botWaypoints = NULL;
		}
		g_numBotWaypoints = 0;
	}
}

void R_ClearBotWP_f(void)
{
	if (!developer.value)
	{
		return;
	}

	if (g_botWaypoints)
	{
		free(g_botWaypoints);
		g_botWaypoints = NULL;
	}
	g_numBotWaypoints = 0;
}

void R_SaveBotWP_f(void)
{
	char wpfilename[MAX_OSPATH];
	char truncbspname[MAX_OSPATH];
	char *p;
	FILE *f;
	int i;

	if (!developer.value || !cl.worldmodel || !g_botWaypoints || g_numBotWaypoints <= 0)
	{
		return;
	}

	strcpy(truncbspname, cl.worldmodel->name);
	p = strstr(truncbspname, ".bsp");
	if (p)
	{
		*p = 0;
	}

	sprintf (wpfilename, "%s/%s.qrb", com_gamedir, truncbspname);
	f = fopen (wpfilename, "w+b");
	if (!f)
	{
		return;
	}
	fwrite(&g_numBotWaypoints, sizeof(g_numBotWaypoints), 1, f);
	fwrite(g_botWaypoints, sizeof(botWaypoint_t), g_numBotWaypoints, f);
	fclose(f);
	Con_Printf("Saved map waypoints.\n");
}

void R_LoadBotWP_f(void)
{
	char wpfilename[MAX_OSPATH];
	char truncbspname[MAX_OSPATH];
	char *p;
	FILE *f;
	int i;

	if (!developer.value || !cl.worldmodel)
	{
		return;
	}

	strcpy(truncbspname, cl.worldmodel->name);
	p = strstr(truncbspname, ".bsp");
	if (p)
	{
		*p = 0;
	}

	sprintf (wpfilename, "%s/%s.qrb", com_gamedir, truncbspname);
	f = fopen (wpfilename, "rb");
	if (!f)
	{
		return;
	}
	fread(&g_numBotWaypoints, sizeof(g_numBotWaypoints), 1, f);
	if (!g_botWaypoints)
	{
		g_botWaypoints = malloc(sizeof(botWaypoint_t)*MAX_BOT_WAYPOINTS);
	}
	fread(g_botWaypoints, sizeof(botWaypoint_t), g_numBotWaypoints, f);
	fclose(f);
	Con_Printf("Loaded map waypoints.\n");
}

/*
===============
R_Init
===============
*/
void R_Init (void)
{	
	R_GammaInit(); //rww

	Cmd_AddCommand ("timerefresh", R_TimeRefresh_f);	
	Cmd_AddCommand ("envmap", R_Envmap_f);	
	Cmd_AddCommand ("pointfile", R_ReadPointFile_f);	
	Cmd_AddCommand("swapinterval", R_SwapInterval_f); //rww

	Cmd_AddCommand("botwpadd", R_AddBotWP_f); //rww
	Cmd_AddCommand("botwprem", R_RemBotWP_f); //rww
	Cmd_AddCommand("botwpclear", R_ClearBotWP_f); //rww
	Cmd_AddCommand("botwpsave", R_SaveBotWP_f); //rww
	Cmd_AddCommand("botwpload", R_LoadBotWP_f); //rww

	Cvar_RegisterVariable (&r_norefresh);
	Cvar_RegisterVariable (&r_lightmap);
	Cvar_RegisterVariable (&r_fullbright);
	Cvar_RegisterVariable (&r_drawentities);
	Cvar_RegisterVariable (&r_drawviewmodel);
	Cvar_RegisterVariable (&r_shadows);
	Cvar_RegisterVariable (&r_mirroralpha);
	Cvar_RegisterVariable (&r_wateralpha);
	Cvar_RegisterVariable (&r_fancywater); //rww
	Cvar_RegisterVariable (&r_dynamic);
	Cvar_RegisterVariable (&r_novis);
	Cvar_RegisterVariable (&r_speeds);
	Cvar_RegisterVariable (&r_netgraph);

	Cvar_RegisterVariable (&gl_clear);
	Cvar_RegisterVariable (&gl_texsort);
 
	//rww - no
 	//if (gl_mtexable)
	//	Cvar_SetValue ("gl_texsort", 0.0);

	Cvar_RegisterVariable (&gl_cull);
	Cvar_RegisterVariable (&gl_smoothmodels);
	Cvar_RegisterVariable (&gl_affinemodels);
	Cvar_RegisterVariable (&gl_polyblend);
	Cvar_RegisterVariable (&gl_flashblend);
	Cvar_RegisterVariable (&gl_playermip);
	Cvar_RegisterVariable (&gl_nocolors);
	Cvar_RegisterVariable (&gl_finish);

	Cvar_RegisterVariable (&gl_fognear); //rww
	Cvar_RegisterVariable (&gl_fogfar); //rww
	Cvar_RegisterVariable (&gl_fog_r); //rww
	Cvar_RegisterVariable (&gl_fog_g); //rww
	Cvar_RegisterVariable (&gl_fog_b); //rww

	Cvar_RegisterVariable (&gl_keeptjunctions);
	Cvar_RegisterVariable (&gl_reporttjunctions);

	if (!g_canUseTexFP)
	{ //rww
		Cvar_SetValue("gl_flashblend", 1.0f); //use silly shell for dlights
		Cvar_SetValue("gl_hq4x", 2.0f); //only use hq2x
	}

	R_InitBubble();
	
	R_InitParticles ();
	R_InitParticleTexture ();

#ifdef GLTEST
	Test_Init ();
#endif

	netgraphtexture = texture_extension_number;
	texture_extension_number++;

	playertextures = texture_extension_number;
	texture_extension_number += MAX_CLIENTS*2; //rww - *2
}

/*
===============
R_TranslatePlayerSkin

Translates a skin texture by the per-player color lookup
===============
*/
void R_TranslatePlayerSkin (int playernum)
{
	int		top, bottom;
	byte	translate[256];
	unsigned	translate32[256];
	unsigned int *filterPix; //rww
	int		i, j;
	byte	*original;
	static unsigned	pixels[/*512*256*/1024*512]; //rww
	unsigned int *out;
	unsigned	scaled_width, scaled_height;
	int			inwidth, inheight;
	int			tinwidth, tinheight;
	byte		*inrow;
	unsigned	frac, fracstep;
	player_info_t *player;
	extern	byte		player_8bit_texels[320*200];
	char s[512];

	GL_DisableMultitexture();

	player = &cl.players[playernum];
	if (!player->name[0])
		return;

	strcpy(s, Info_ValueForKey(player->userinfo, "skin"));
	COM_StripExtension(s, s);
	if (player->skin && !stricmp(s, player->skin->name))
		player->skin = NULL;

	if (player->_topcolor != player->topcolor ||
		player->_bottomcolor != player->bottomcolor || !player->skin) {
		player->_topcolor = player->topcolor;
		player->_bottomcolor = player->bottomcolor;

		top = player->topcolor;
		bottom = player->bottomcolor;
		top = (top < 0) ? 0 : ((top > 13) ? 13 : top);
		bottom = (bottom < 0) ? 0 : ((bottom > 13) ? 13 : bottom);
		top *= 16;
		bottom *= 16;

		for (i=0 ; i<256 ; i++)
			translate[i] = i;

		for (i=0 ; i<16 ; i++)
		{
			if (top < 128)	// the artists made some backwards ranges.  sigh.
				translate[TOP_RANGE+i] = top+i;
			else
				translate[TOP_RANGE+i] = top+15-i;
					
			if (bottom < 128)
				translate[BOTTOM_RANGE+i] = bottom+i;
			else
				translate[BOTTOM_RANGE+i] = bottom+15-i;
		}

		//
		// locate the original skin pixels
		//
		// real model width
		tinwidth = 296;
		tinheight = 194;

		if (!player->skin)
			Skin_Find(player);
		if ((original = Skin_Cache(player->skin)) != NULL) {
			//skin data width
			inwidth = 320;
			inheight = 200;
		} else {
			original = player_8bit_texels;
			inwidth = 296;
			inheight = 194;
		}


		// because this happens during gameplay, do it fast
		// instead of sending it through gl_upload 8
		GL_Bind(playertextures + (playernum*2)); //rww - *2

	#if 0
		s = 320*200;
		byte	translated[320*200];

		for (i=0 ; i<s ; i+=4)
		{
			translated[i] = translate[original[i]];
			translated[i+1] = translate[original[i+1]];
			translated[i+2] = translate[original[i+2]];
			translated[i+3] = translate[original[i+3]];
		}


		// don't mipmap these, because it takes too long
		GL_Upload8 (translated, paliashdr->skinwidth, paliashdr->skinheight, 
			false, false, true);
	#endif

		for (i=0 ; i<256 ; i++)
			translate32[i] = d_8to24table[translate[i]];

		//rww - filter and resize instead
#if 1
		scaled_width = gl_max_size.value < 512 ? gl_max_size.value : 512;
		scaled_height = gl_max_size.value < 256 ? gl_max_size.value : 256;
		// allow users to crunch sizes down even more if they want
		scaled_width >>= (int)gl_playermip.value;
		scaled_height >>= (int)gl_playermip.value;

		out = pixels;
		memset(pixels, 0, sizeof(pixels));
		fracstep = tinwidth*0x10000/scaled_width;
		for (i=0 ; i<scaled_height ; i++, out += scaled_width)
		{
			inrow = original + inwidth*(i*tinheight/scaled_height);
			frac = fracstep >> 1;
			for (j=0 ; j<scaled_width ; j+=4)
			{
				out[j] = translate32[inrow[frac>>16]];
				frac += fracstep;
				out[j+1] = translate32[inrow[frac>>16]];
				frac += fracstep;
				out[j+2] = translate32[inrow[frac>>16]];
				frac += fracstep;
				out[j+3] = translate32[inrow[frac>>16]];
				frac += fracstep;
			}
		}

		filterPix = GL_FilterTexture((unsigned char *)pixels, &scaled_width, &scaled_height, false, false, false, NULL);
#else
		scaled_width = gl_max_size.value < 1024 ? gl_max_size.value : 1024;
		scaled_height = gl_max_size.value < 512 ? gl_max_size.value : 512;
		// allow users to crunch sizes down even more if they want
		scaled_width >>= (int)gl_playermip.value;
		scaled_height >>= (int)gl_playermip.value;

		filterPix = GL_FilterTexture(original, &inwidth, &inheight, false, false, true, translate32);
		GL_ResampleTexture (filterPix, inwidth, inheight, pixels, scaled_width, scaled_height);
		free(filterPix);
#endif

		glTexImage2D (GL_TEXTURE_2D, 0, gl_solid_format, 
			scaled_width, scaled_height, 0, GL_RGBA, 
			GL_UNSIGNED_BYTE, filterPix);

		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

		if (g_canUseTexFP)
		{ //rww - create/upload corresponding normal map
			char name[256];
			if (player->skin->failedload)
			{
				strcpy(name, "plskin_base");
			}
			else
			{
				sprintf(name, "plskin_%s", player->skin->name);
			}
			GL_UploadNormal(name, playertextures + (playernum*2), playertextures + (playernum*2) + 1,
				filterPix, scaled_width, scaled_height, false);
		}

		free(filterPix);
	}
}

/*
===============
R_NewMap
===============
*/
int hr_automap = 0;
qboolean g_hasAutomap = false;
void R_NewMap (void)
{
	int		i;
	char	*p; //rww
	
	for (i=0 ; i<256 ; i++)
		d_lightstylevalue[i] = 264;		// normal light value

	memset (&r_worldentity, 0, sizeof(r_worldentity));
	r_worldentity.model = cl.worldmodel;

	//rww
	g_hasAutomap = false;
	p = strstr(cl.worldmodel->name, ".bsp");
	if (p)
	{
		char *to = p;
		char amappath[1024];
		int i;
		if (!hr_automap)
		{
			hr_automap = texture_extension_number;
			texture_extension_number++;
		}
		while (p > cl.worldmodel->name && *p != '/' && *p != '\\')
		{
			p--;
		}
		if (*p == '/' || *p == '\\')
		{
			p++;
		}
		strcpy(amappath, "qdix/amaps/");
		i = strlen(amappath);
		while (p < to)
		{
			amappath[i] = *p;
			i++;
			p++;
		}
		amappath[i] = 0;
		strcat(amappath, ".png");

		if (GL_UploadPNG(amappath, hr_automap, false))
		{
			g_hasAutomap = true;
		}
	}

// clear out efrags in case the level hasn't been reloaded
// FIXME: is this one short?
	for (i=0 ; i<cl.worldmodel->numleafs ; i++)
		cl.worldmodel->leafs[i].efrags = NULL;
		 	
	r_viewleaf = NULL;
	R_ClearParticles ();

	GL_BuildLightmaps ();

	// identify sky texture
	skytexturenum = -1;
	mirrortexturenum = -1;
	for (i=0 ; i<cl.worldmodel->numtextures ; i++)
	{
		if (!cl.worldmodel->textures[i])
			continue;
		if (!Q_strncmp(cl.worldmodel->textures[i]->name,"sky",3) )
			skytexturenum = i;
		if (!Q_strncmp(cl.worldmodel->textures[i]->name,"window02_1",10) )
			mirrortexturenum = i;
 		cl.worldmodel->textures[i]->texturechain = NULL;
	}
#ifdef QUAKE2
	R_LoadSkys ();
#endif
}


/*
====================
R_TimeRefresh_f

For program optimization
====================
*/
void R_TimeRefresh_f (void)
{
	int			i;
	float		start, stop, time;

	glDrawBuffer  (GL_FRONT);
	glFinish ();

	start = Sys_DoubleTime ();
	for (i=0 ; i<128 ; i++)
	{
		r_refdef.viewangles[1] = i/128.0*360.0;
		R_RenderView ();
	}

	glFinish ();
	stop = Sys_DoubleTime ();
	time = stop-start;
	Con_Printf ("%f seconds (%f fps)\n", time, 128/time);

	glDrawBuffer  (GL_BACK);
	GL_EndRendering ();
}

void D_FlushCaches (void)
{
}


